# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the utilities."""
import json
from inspect import signature
from typing import Awaitable
from typing import Callable
from typing import cast

import structlog
from aio_pika import IncomingMessage

from .metrics import exception_parse_counter


logger = structlog.get_logger()


CallbackType = Callable[[IncomingMessage], Awaitable]


def function_to_name(function: Callable) -> str:
    """Get a uniquely qualified name for a given function."""
    return function.__name__


def decode_body(message: IncomingMessage) -> str:
    """Decode the message body into a string.

    Args:
        message: The aio_pika message.

    Returns:
        The decoded body as a string.
    """
    routing_key = message.routing_key
    with exception_parse_counter.labels(routing_key).count_exceptions():
        decoded = message.body.decode("utf-8")
    logger.debug("Decoded body", routing_key=routing_key, decoded=decoded)
    return decoded


def parse_payload(message: IncomingMessage) -> dict:
    """Parse the message body as json.

    Args:
        message: The aio_pika message.

    Returns:
        JSON parsed body.
    """
    routing_key = message.routing_key
    with exception_parse_counter.labels(routing_key).count_exceptions():
        decoded = decode_body(message)
        payload = json.loads(decoded)
    logger.debug("Parsed body", routing_key=routing_key, payload=payload)
    return cast(dict, payload)


# Map af parameters supported by pass_arguments
parameter_map = {
    "body": decode_body,
    "message": lambda message: message,
    "message_id": lambda message: message.message_id,
    "payload": parse_payload,
    "routing_key": lambda message: message.routing_key,
}


def pass_arguments(function: Callable) -> CallbackType:
    """Decorator which automatically transform parameters as desired.

    The decorator inspects the parameters taken by the provided function and
    transforms the AMQP message to fit those parameters, then calls the function
    with the requested parameters.

    Examples:
        We may wish to just receive the callback without any arguments:
        ```
        @amqp_system.register("my.routing.key")
        @pass_arguments
        async def callback() -> None:
            pass
        ```
        Or we may wish to only receive the `routing_key`:
        ```
        @amqp_system.register("my.routing.key")
        @amqp_system.register("your.routing.key")
        @pass_arguments
        async def callback(routing_key: str) -> None:
            pass
        ```
        Or the `message_id` and the `payload`:
        ```
        @amqp_system.register("my.routing.key")
        @pass_arguments
        async def callback(message_id: str, payload: dict) -> None:
            pass
        ```
        Alternatively without the decorator we receive the `routing_key` and `message`:
        ```
        @amqp_system.register("my.routing.key")
        async def callback(routing_key: str, message: IncomingMessage) -> None:
            pass
        ```

    Note:
        The parameters that can be mapped to are found in the `parameter_map`.

    Args:
        function: The callable to match parameters to.

    Returns:
        Callable with the signature expected by AMQPSystem.register
    """
    sig = signature(function)
    parameters = list(sig.parameters.keys())

    # Verify that all parameters of the function can be mapped
    # Having this outside the decorator function allows earlier errors
    for parameter in parameters:
        assert parameter in parameter_map, "Unknown parameter in pass_argument"

    # Function mapping parameters from CallbackType to function
    async def message_receiver(message: IncomingMessage) -> None:
        """Maps parameters via the parameter_map and calls the decorated function.

        Note:
            This function fulfills the CallbackType signature, thus the decorator
            maps from arbitrary function signatures to the CallbackType signature.

        Args:
            message: The aio_pika message.

        Returns:
            None
        """
        kwargs = {}
        for parameter in parameters:
            kwargs[parameter] = parameter_map[parameter](message)
        await function(**kwargs)

    return message_receiver



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the Generic AMQPSystem."""
from .abstract_amqpsystem import AbstractAMQPSystem


class AMQPSystem(AbstractAMQPSystem):
    """Generic AMQPSystem."""

    register = AbstractAMQPSystem._register
    publish_message = AbstractAMQPSystem._publish_message



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the Abstract AMQPSystem."""
import asyncio
import json
from abc import ABCMeta
from contextlib import contextmanager
from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Set

import structlog
from aio_pika import connect_robust
from aio_pika import ExchangeType
from aio_pika import IncomingMessage
from aio_pika import Message
from aio_pika.abc import AbstractChannel
from aio_pika.abc import AbstractExchange
from aio_pika.abc import AbstractQueue
from aio_pika.abc import AbstractRobustConnection
from more_itertools import all_unique

from .config import ConnectionSettings
from .metrics import backlog_count
from .metrics import callbacks_registered
from .metrics import exception_callback_counter
from .metrics import last_loop_periodic
from .metrics import last_on_message
from .metrics import last_periodic
from .metrics import processing_calls
from .metrics import processing_inprogress
from .metrics import processing_time
from .metrics import routes_bound
from .utils import CallbackType
from .utils import function_to_name


logger = structlog.get_logger()


@contextmanager
def _handle_metrics(
    routing_key: str, function_name: str
) -> Generator[None, None, None]:
    """Expose metrics about a callback.

    Args:
        labels: Labels to apply to the metrics.

    Returns:
        None
    """
    labels = (routing_key, function_name)

    processing_calls.labels(*labels).inc()
    with exception_callback_counter.labels(*labels).count_exceptions():
        with processing_inprogress.labels(*labels).track_inprogress():
            with processing_time.labels(*labels).time():
                yield


async def _on_message(callback: CallbackType, message: IncomingMessage) -> None:
    """AbstractAMQPSystem message handler.

    Handles logging, metrics, retrying and exception handling.

    Args:
        callback: The callback to call with the message.
        message: The message to deliver to the callback.

    Returns:
        None
    """
    last_on_message.set_to_current_time()

    assert message.routing_key is not None
    routing_key = message.routing_key
    function_name = function_to_name(callback)
    log = logger.bind(function=function_name, routing_key=routing_key)

    log.debug("Received message")
    try:
        with _handle_metrics(routing_key, function_name):
            # Requeue messages on exceptions, so they can be retried.
            async with message.process(requeue=True):
                await callback(message)
    except Exception as exception:
        log.exception("Exception during on_message()")
        raise exception


class AbstractAMQPSystem:
    """Abstract base-class for AMQPSystems.

    Shared code used by both AMQPSystem and MOAMQPSystem.
    """

    __metaclass__ = ABCMeta

    def __init__(self) -> None:
        self._registry: Dict[CallbackType, Set[str]] = {}

        self._connection: Optional[AbstractRobustConnection] = None
        self._channel: Optional[AbstractChannel] = None
        self._exchange: Optional[AbstractExchange] = None

        self._periodic_task: Optional[asyncio.Task] = None

    @property
    def started(self) -> bool:
        """Whether a connection has been made.

        Returns:
            Whether a connection has been made.
        """
        return self._connection is not None

    def _setup_periodic_metrics(self, queues: Dict[str, AbstractQueue]) -> None:
        """Setup a periodic job to update non-eventful metrics.

        Args:
            queues: The aio_pika queues to export periodic metrics for.

        Returns:
            None
        """
        # Setup periodic metrics
        async def periodic_metrics() -> None:
            loop = asyncio.get_running_loop()
            while True:
                last_periodic.set_to_current_time()
                last_loop_periodic.set(loop.time())
                for function_name, queue in queues.items():
                    backlog_count.labels(function_name).set(
                        queue.declaration_result.message_count
                    )
                await asyncio.sleep(1)

        self._periodic_task = asyncio.create_task(periodic_metrics())

    async def start(self, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        """Start the AMQPSystem.

        This method:
        * Connects to the AMQP server.
        * Sets up the AMQP exchange if it does not exist
        * Creates durable queues for each callback in the callback registry.
        * Binds routes to the queues according to the callback registry.
        * Setups up periodic metrics.

        Args:
            *args: Arguments are forwarded directly to ConnectionSettings.
            **kwargs: Arguments are forwarded directly to ConnectionSettings.

        Returns:
            None
        """
        settings = ConnectionSettings(*args, **kwargs)

        # We expect no active connections to be established
        assert self._connection is None
        assert self._channel is None
        assert self._exchange is None

        # We expect function_to_name to be unique for each callback
        assert all_unique(map(function_to_name, self._registry.keys()))

        # We except queue_prefix to be set if any callbacks are registered
        if self._registry:
            assert settings.queue_prefix is not None

        logger.info(
            "Establishing AMQP connection",
            scheme=settings.amqp_url.scheme,
            user=settings.amqp_url.user,
            host=settings.amqp_url.host,
            port=settings.amqp_url.port,
            path=settings.amqp_url.path,
        )
        self._connection = await connect_robust(settings.amqp_url)

        logger.info("Creating AMQP channel")
        self._channel = await self._connection.channel()
        await self._channel.set_qos(prefetch_count=10)

        logger.info(
            "Attaching AMQP exchange to channel", exchange=settings.amqp_exchange
        )
        # Make our exchange durable so it survives broker restarts
        self._exchange = await self._channel.declare_exchange(
            settings.amqp_exchange, ExchangeType.TOPIC, durable=True
        )

        # TODO: Create queues and binds in parallel?
        queues = {}
        for callback, routing_keys in self._registry.items():
            function_name = function_to_name(callback)
            log = logger.bind(function=function_name)

            queue_name = f"{settings.queue_prefix}_{function_name}"
            log.info("Declaring unique message queue", queue_name=queue_name)
            # Make our queues durable so they survive broker restarts
            queue = await self._channel.declare_queue(queue_name, durable=True)
            queues[function_name] = queue

            log.info("Starting message listener")
            await queue.consume(partial(_on_message, callback))  # type: ignore

            log.info("Binding routing keys")
            for routing_key in routing_keys:
                log.info("Binding routing-key", routing_key=routing_key)
                await queue.bind(self._exchange, routing_key=routing_key)
                routes_bound.labels(function_name).inc()

        self._setup_periodic_metrics(queues)

    async def stop(self) -> None:
        """Stop the AMQPSystem.

        This method disconnects from the AMQP server, and stops the periodic metrics.

        Returns:
            None
        """
        if self._periodic_task is not None:
            self._periodic_task.cancel()
            self._periodic_task = None

        self._exchange = None

        if self._channel is not None:
            await self._channel.close()
            self._channel = None

        if self._connection is not None:
            await self._connection.close()
            self._connection = None

    def run_forever(self, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        """Start the AMQPSystem and run it forever.

        Args:
            *args: Arguments are forwarded directly to ConnectionSettings.
            **kwargs: Arguments are forwarded directly to ConnectionSettings.

        Returns:
            None
        """
        loop = asyncio.get_event_loop()
        # Setup everything
        loop.run_until_complete(self.start(*args, **kwargs))
        # Run forever listening to messages
        loop.run_forever()
        # Clean up everything
        loop.run_until_complete(self.stop())
        loop.close()

    def _register(self, routing_key: str) -> Callable[[CallbackType], CallbackType]:
        """Get a decorator for registering callbacks.

        Examples:
            ```
            address_create_decorator = amqp_system.register("EMPLOYEE.ADDRESS.CREATE")

            @address_create_decorator
            def callback1(message: IncomingMessage):
                pass

            @address_create_decorator
            def callback2(message: IncomingMessage):
                pass
            ```
            Or directly:
            ```
            @amqp_system.register("EMPLOYEE.ADDRESS.CREATE")
            def callback1(message: IncomingMessage):
                pass

            @amqp_system.register("EMPLOYEE.ADDRESS.CREATE")
            def callback2(message: IncomingMessage):
                pass
            ```

        Args:
            routing_key: The routing key to bind messages for.

        Returns:
            A decorator for registering a function to receive callbacks.
        """
        assert routing_key != ""

        def decorator(function: CallbackType) -> CallbackType:
            """Registers the given callback and routing key in the callback registry.

            Args:
                function: The callback to registry.

            Returns:
                The unmodified given function.
            """
            function_name = function_to_name(function)

            log = logger.bind(routing_key=routing_key, function=function_name)
            log.info("Register called")

            if self.started:
                message = "Cannot register callback after run() has been called!"
                log.error(message)
                raise ValueError(message)

            callbacks_registered.labels(routing_key).inc()
            self._registry.setdefault(function, set()).add(routing_key)
            return function

        return decorator

    async def _publish_message(self, routing_key: str, payload: dict) -> None:
        """Publish a message to the given routing key.

        Args:
            routing_key: The routing key to send the message to.
            payload: The message payload.

        Returns:
            None
        """
        if self._exchange is None:
            raise ValueError("Must call start() before publish message!")

        message = Message(body=json.dumps(payload).encode("utf-8"))
        await self._exchange.publish(routing_key=routing_key, message=message)



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the all pydantic BaseSetting(s)."""
from typing import Optional

from pydantic import AmqpDsn
from pydantic import BaseSettings
from pydantic import parse_obj_as


class ConnectionSettings(BaseSettings):
    """Settings for the AMQP connection.

    These settings can either be provided via environmental variables or as overrides
    to the AMQPSystem.start method via. args / kwargs.

    Attributes:
        queue_prefix:
            Program specific queue name prefix, should be globally unique, but
            consistent across program restarts. The program name is a good candidate.

            Only required for receiving.

        amqp_url:
            The AMQP connection url, including credentials and vhost.

        amqp_exchange:
            The AMQP Exchange we are binding queue messages from. Default is os2mo,
            but can be overridden for integration specific exchanges.
    """

    queue_prefix: Optional[str]
    amqp_url: AmqpDsn = parse_obj_as(AmqpDsn, "amqp://guest:guest@localhost:5672")
    amqp_exchange: str = "os2mo"



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the MO specific AMQP models."""
from datetime import datetime
from enum import auto
from enum import Enum
from enum import unique
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import cast
from typing import List
from uuid import UUID

from pydantic import BaseModel


class AutoNameEnum(Enum):
    """Enum subclass which lets `auto` assign the enum key as value.

    From: https://docs.python.org/3/library/enum.html#using-automatic-values

    Example:
        ```
        class Color(Enum):
            RED = auto()
            BLUE = auto()

        print(list(Color))
        ```
        Yields:
        ```
        [<Color.RED: 1>, <Color.BLUE: 2>]
        ```
        Whereas:
        ```
        class Color(str, AutoNameEnum):
            RED = auto()
            BLUE = auto()

        print(list(Color))
        ```
        Yields:
        ```
        [<Color.RED: 'RED'>, <Color.BLUE: 'BLUE'>]
        ```
    """

    @staticmethod
    def _generate_next_value_(
        name: Any, start: int, count: int, last_values: List[Any]
    ) -> str:
        # The next `auto` value is simply the key name
        return cast(str, name)


@unique
class ServiceType(str, AutoNameEnum):
    """The MO service types.

    Describes the root object for the object the operation was executed on.
    """

    EMPLOYEE: str = cast(str, auto())
    ORG_UNIT: str = cast(str, auto())
    WILDCARD: str = "*"


@unique
class ObjectType(str, AutoNameEnum):
    """The MO object types.

    Describes the object the operation was executed on.
    """

    ADDRESS: str = cast(str, auto())
    ASSOCIATION: str = cast(str, auto())
    EMPLOYEE: str = cast(str, auto())
    ENGAGEMENT: str = cast(str, auto())
    IT: str = cast(str, auto())
    KLE: str = cast(str, auto())
    LEAVE: str = cast(str, auto())
    MANAGER: str = cast(str, auto())
    OWNER: str = cast(str, auto())
    ORG_UNIT: str = cast(str, auto())
    RELATED_UNIT: str = cast(str, auto())
    ROLE: str = cast(str, auto())
    WILDCARD: str = "*"


@unique
class RequestType(str, AutoNameEnum):
    """The MO request types.

    Describes the type of operation which was executed.

    Args:
        CREATE: Something was created.
        EDIT: Something was edited.
        TERMINATE: Something was terminated.
        REFRESH: Something was requested to be refreshed.
        WILDCARD: Something happened.
    """

    CREATE: str = cast(str, auto())
    EDIT: str = cast(str, auto())
    TERMINATE: str = cast(str, auto())
    REFRESH: str = cast(str, auto())
    WILDCARD: str = "*"


class PayloadType(BaseModel):
    """MO AMQP message format.

    Args:
        uuid: The UUID of the org-unit or employee.
        object_uuid: The UUID of the object itself.
        time: Time when MO emitted the message.
    """

    uuid: UUID
    object_uuid: UUID
    time: datetime


MOCallbackType = Callable[
    [ServiceType, ObjectType, RequestType, PayloadType], Awaitable
]



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the MO specific AMQPSystem."""
from typing import Callable
from typing import Tuple

from aio_pika import IncomingMessage
from fastapi.encoders import jsonable_encoder
from pydantic import parse_raw_as

from .abstract_amqpsystem import AbstractAMQPSystem
from .mo_models import MOCallbackType
from .mo_models import ObjectType
from .mo_models import PayloadType
from .mo_models import RequestType
from .mo_models import ServiceType


MORoutingTuple = Tuple[ServiceType, ObjectType, RequestType]


def to_routing_key(
    service_type: ServiceType, object_type: ObjectType, request_type: RequestType
) -> str:
    """Convert the given service.object.request tuple to a routing_key.

    Args:
        service_type: The service type to convert.
        object_type: The object type to convert.
        request_type: The request type to convert.

    Returns:
        The equivalent routing key.
    """
    return f"{service_type}.{object_type}.{request_type}"


def from_routing_key(routing_key: str) -> MORoutingTuple:
    """Convert the given routing_key to a service.object.request tuple.

    Raises:
        ValueError: Raised if the routing_key does not contain 2 periods.
        ValueError: Raised if the routing_key components cannot be parsed.

    Args:
        routing_key: The routing key to convert.

    Returns:
        The equivalent service.object.request tuple.
    """
    parts = routing_key.split(".")
    if len(parts) != 3:
        raise ValueError("Expected a three tuple!")
    service_str, object_str, request_str = parts
    return ServiceType(service_str), ObjectType(object_str), RequestType(request_str)


class MOAMQPSystem(AbstractAMQPSystem):
    """MO specific AMQPSystem.

    Has specifically tailored `register` and `publish_message` methods.
    Both of which utilize the MO AMQP routing-key structure and payload format.
    """

    def register(
        self,
        service_type: ServiceType,
        object_type: ObjectType,
        request_type: RequestType,
    ) -> Callable:
        """Get a decorator for registering callbacks.

        Examples:
            ```
            address_create_decorator = moamqp_system.register(
                ServiceType.EMPLOYEE,
                ObjectType.ADDRESS,
                RequestType.CREATE
            )
            @address_create_decorator
            def callback(
                service_type: ServiceType,
                object_type: ObjectType,
                request_type: RequestType,
                payload: PayloadType
            ) -> None:
                pass
            ```
            Or directly:
            ```
            @moamqp_system.register(
                ServiceType.EMPLOYEE,
                ObjectType.ADDRESS,
                RequestType.CREATE
            )
            def callback(
                service_type: ServiceType,
                object_type: ObjectType,
                request_type: RequestType,
                payload: PayloadType
            ) -> None:
                pass
            ```


        Args:
            service_type: The service type to bind messages for.
            object_type: The object type to bind messages for.
            request_type: The request type to bind messages for.

        Returns:
            A decorator for registering a function to receive callbacks.
        """
        routing_key = to_routing_key(service_type, object_type, request_type)
        amqp_decorator = self._register(routing_key)

        def decorator(function: MOCallbackType) -> MOCallbackType:
            async def wrapper(message: IncomingMessage) -> None:
                assert message.routing_key is not None
                routing_tuple = from_routing_key(message.routing_key)
                payload = parse_raw_as(PayloadType, message.body)
                await function(*routing_tuple, payload)

            wrapper.__name__ = function.__name__
            amqp_decorator(wrapper)
            return function

        return decorator

    async def publish_message(
        self,
        service_type: ServiceType,
        object_type: ObjectType,
        request_type: RequestType,
        payload: PayloadType,
    ) -> None:
        """Publish a message to the given service.object.request tuple.

        Args:
            service_type: The service type to send the message to.
            object_type: The object type to send the message to.
            request_type: The request type to send the message to.
            payload: The message payload.

        Returns:
            None
        """
        routing_key = to_routing_key(service_type, object_type, request_type)
        payload_obj = jsonable_encoder(payload)
        await self._publish_message(routing_key, payload_obj)



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the all prometheus metrics."""
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

exception_callback_counter = Counter(
    "amqp_exceptions_callback",
    "Exception counter",
    ["routing_key", "function"],
)
exception_parse_counter = Counter(
    "amqp_exceptions_parse",
    "Exception counter",
    ["routing_key"],
)
processing_time = Histogram(
    "amqp_processing_seconds",
    "Time spent running callback",
    ["routing_key", "function"],
)
processing_inprogress = Gauge(
    "amqp_inprogress",
    "Number of callbacks currently running",
    ["routing_key", "function"],
)
processing_calls = Counter(
    "amqp_calls",
    "Number of callbacks made",
    ["routing_key", "function"],
)
last_on_message = Gauge("amqp_last_on_message", "Timestamp of the last on_message call")
last_periodic = Gauge("amqp_last_periodic", "Timestamp of the last periodic call")
last_loop_periodic = Gauge(
    "amqp_last_loop_periodic", "Timestamp (monotonic) of the last periodic call"
)
backlog_count = Gauge(
    "amqp_backlog",
    "Number of messages waiting for processing in the backlog",
    ["function"],
)
routes_bound = Counter(
    "amqp_routes_bound", "Number of routing-keys bound to the queue", ["function"]
)
callbacks_registered = Counter(
    "amqp_callbacks_registered", "Number of callbacks registered", ["routing_key"]
)
